"""Tests for aiopvpc."""

from datetime import datetime, timedelta
from typing import cast

import pytest

from aiopvpc.const import (
    ALL_SENSORS,
    DataSource,
    KEY_ADJUSTMENT,
    KEY_INJECTION,
    KEY_MAG,
    KEY_OMIE,
    KEY_PVPC,
    REFERENCE_TZ,
    SENSOR_KEY_TO_DATAID,
)
from aiopvpc.pvpc_data import PVPCData
from tests.conftest import MockAsyncSession, TZ_TEST


@pytest.mark.parametrize(
    "ts, source, timezone, n_prices, n_calls, n_prices_8h, available_8h",
    (
        ("2021-06-01 09:00:00", "esios", REFERENCE_TZ, 24, 1, 24, True),
        ("2021-06-01 09:00:00", "esios", TZ_TEST, 24, 1, 24, True),
        ("2024-03-09 09:00:00", "esios", REFERENCE_TZ, 24, 1, 24, True),
        ("2024-03-09 09:00:00", "esios", TZ_TEST, 24, 1, 24, True),
        ("2021-10-30 00:00:00+08:00", "esios_public", TZ_TEST, 0, 1, 0, False),
        ("2021-10-30 00:00:00", "esios_public", TZ_TEST, 24, 1, 24, True),
        ("2021-10-31 00:00:00", "esios_public", TZ_TEST, 25, 1, 25, True),
        ("2022-03-27 20:00:00", "esios_public", TZ_TEST, 23, 2, 23, False),
        ("2022-03-27 20:00:00+04:00", "esios_public", TZ_TEST, 23, 1, 23, False),
        ("2021-10-30 21:00:00", "esios_public", TZ_TEST, 49, 2, 26, True),
        ("2021-10-30 21:00:00+01:00", "esios_public", TZ_TEST, 49, 2, 26, True),
        ("2021-10-30 00:00:00", "esios_public", REFERENCE_TZ, 24, 1, 24, True),
        ("2021-10-31 00:00:00", "esios_public", REFERENCE_TZ, 25, 1, 25, True),
        ("2022-03-27 20:00:00", "esios_public", REFERENCE_TZ, 23, 2, 23, False),
        ("2021-10-30 21:00:00", "esios_public", REFERENCE_TZ, 49, 2, 25, True),
        ("2021-06-01 09:00:00", "esios_public", REFERENCE_TZ, 24, 1, 24, True),
        ("2021-06-01 09:00:00", "esios_public", TZ_TEST, 24, 1, 24, True),
    ),
)
@pytest.mark.asyncio
async def test_price_extract(
    ts,
    source,
    timezone,
    n_prices,
    n_calls,
    n_prices_8h,
    available_8h,
):
    """Test data parsing of official API files."""
    day = datetime.fromisoformat(ts)
    mock_session = MockAsyncSession()

    pvpc_data = PVPCData(
        session=mock_session,
        tariff="2.0TD",
        local_timezone=timezone,
        data_source=cast(DataSource, source),
        api_token="test-token" if source == "esios" else None,
    )

    api_data = await pvpc_data.async_update_all(None, day)
    pvpc_data.process_state_and_attributes(api_data, KEY_PVPC, day)
    assert len(api_data.sensors[KEY_PVPC]) == n_prices
    assert mock_session.call_count == n_calls
    assert len(api_data.sensors) == 1

    has_prices = pvpc_data.process_state_and_attributes(
        api_data, KEY_PVPC, day + timedelta(hours=10)
    )
    assert len(api_data.sensors[KEY_PVPC]) == n_prices_8h
    assert has_prices == available_8h
    if has_prices:
        last_dt, last_p = list(api_data.sensors[KEY_PVPC].items())[-1]
        assert last_dt.astimezone(timezone).hour == 23


@pytest.mark.asyncio
async def test_price_sensor_attributes():
    """Test data parsing of official API files."""
    day = datetime.fromisoformat("2024-03-09 09:00:00")
    mock_session = MockAsyncSession()

    pvpc_data = PVPCData(
        session=mock_session,
        tariff="2.0TD",
        api_token="test-token",
        sensor_keys=ALL_SENSORS,
    )

    api_data = await pvpc_data.async_update_all(None, day)
    for key in ALL_SENSORS:
        pvpc_data.process_state_and_attributes(api_data, key, day)
    assert len(api_data.sensors[KEY_PVPC]) == 24
    assert mock_session.call_count == 5
    assert len(api_data.sensors) == 6

    ref_data = {
        KEY_PVPC: {"hours_to_better_price": 1, "num_better_prices_ahead": 3},
        KEY_INJECTION: {"hours_to_better_price": 6, "num_better_prices_ahead": 6},
        KEY_MAG: {"hours_to_better_price": 1, "num_better_prices_ahead": 11},
        KEY_OMIE: {"hours_to_better_price": 1, "num_better_prices_ahead": 4},
        KEY_ADJUSTMENT: {"hours_to_better_price": 1, "num_better_prices_ahead": 2},
    }

    for key in ALL_SENSORS:
        has_prices = pvpc_data.process_state_and_attributes(
            api_data, key, day + timedelta(hours=2)
        )
        assert has_prices, key
        assert api_data.availability[key]
        last_dt, last_p = list(api_data.sensors[key].items())[-1]
        assert last_dt.astimezone(REFERENCE_TZ).hour == 23

        current_price = pvpc_data.states[key]
        sensor_attrs = pvpc_data.sensor_attributes[key]
        assert sensor_attrs["sensor_id"] == key
        assert sensor_attrs["data_id"] == SENSOR_KEY_TO_DATAID.get(key, "composed")
        assert sensor_attrs["price_12h"] == current_price
        prices_ahead = [sensor_attrs[f"price_{h:02}h"] for h in range(13, 24)]
        assert len(prices_ahead) == 11
        assert sensor_attrs["price_23h"] == last_p
        assert sensor_attrs["min_price"] == min(api_data.sensors[key].values())
        assert sensor_attrs["max_price"] == max(api_data.sensors[key].values())
        key_min_at = f'price_{sensor_attrs["min_price_at"]:02d}h'
        assert sensor_attrs[key_min_at] == min(api_data.sensors[key].values())
        key_max_at = f'price_{sensor_attrs["max_price_at"]:02d}h'
        assert sensor_attrs[key_max_at] == max(api_data.sensors[key].values())
        assert (
            sensor_attrs["hours_to_better_price"]
            == ref_data[key]["hours_to_better_price"]
        )
        assert (
            sensor_attrs["num_better_prices_ahead"]
            == ref_data[key]["num_better_prices_ahead"]
        )
        key_next = f'price_{12 + sensor_attrs["hours_to_better_price"]}h'
        if key == KEY_INJECTION:
            assert sensor_attrs[key_next] > current_price
            num_better = sum(1 for p in prices_ahead if p > current_price)
        else:
            assert sensor_attrs[key_next] < current_price
            num_better = sum(1 for p in prices_ahead if p < current_price)
        assert num_better == sensor_attrs["num_better_prices_ahead"]



"""Tests for aiopvpc."""

from datetime import datetime
from typing import cast

import pytest

from aiopvpc.const import DataSource, KEY_PVPC, REFERENCE_TZ, TARIFFS, UTC_TZ
from aiopvpc.pvpc_data import PVPCData
from tests.conftest import MockAsyncSession, TZ_TEST


@pytest.mark.parametrize(
    "local_tz, source, tariff, expected_18h",
    (
        (TZ_TEST, "esios_public", TARIFFS[0], 0.23144),
        (REFERENCE_TZ, "esios_public", TARIFFS[0], 0.23144),
        (REFERENCE_TZ, "esios_public", TARIFFS[1], 0.13813),
        (REFERENCE_TZ, "esios", TARIFFS[0], 0.23144),
        (REFERENCE_TZ, "esios", TARIFFS[1], 0.13813),
        (TZ_TEST, "esios", TARIFFS[0], 0.23144),
    ),
)
@pytest.mark.asyncio
async def test_geo_ids(local_tz, source, tariff, expected_18h):
    """Test different prices for different geo zones."""
    start = datetime(2021, 6, 1, 10, tzinfo=UTC_TZ)
    mock_session = MockAsyncSession()
    pvpc_data = PVPCData(
        session=mock_session,
        tariff=tariff,
        local_timezone=local_tz,
        data_source=cast(DataSource, source),
        api_token="test-token" if source == "esios" else None,
    )
    api_data = await pvpc_data.async_update_all(None, start)
    assert all(api_data.availability.values())
    assert pvpc_data.process_state_and_attributes(api_data, KEY_PVPC, start)
    # for ts, price in pvpc_data._current_prices.items():
    #     print(f"{ts.astimezone(local_tz):%H}h --> {price:.5f} ")
    ts_loc_18h_utc = datetime(2021, 6, 1, 18, tzinfo=local_tz).astimezone(UTC_TZ)
    price_loc_18h = api_data.sensors[KEY_PVPC][ts_loc_18h_utc]
    assert price_loc_18h == expected_18h



"""Tests for aiopvpc."""

from __future__ import annotations

import json
import logging
import pathlib
import zoneinfo
from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING

from aiopvpc.const import (
    ESIOS_INJECTION,
    ESIOS_MAG,
    ESIOS_MARKET_ADJUSTMENT,
    ESIOS_OMIE,
    ESIOS_PVPC,
    KEY_PVPC,
)

if TYPE_CHECKING:
    from aiopvpc.pvpc_data import EsiosApiData, PVPCData

TEST_EXAMPLES_PATH = pathlib.Path(__file__).parent / "api_examples"
TZ_TEST = zoneinfo.ZoneInfo("Atlantic/Canary")

_FIXTURE_DATA_2021_10_30 = "PVPC_CURV_DD_2021_10_30.json"
_FIXTURE_DATA_2021_10_31 = "PVPC_CURV_DD_2021_10_31.json"
_FIXTURE_DATA_2022_03_27 = "PVPC_CURV_DD_2022_03_27.json"
_FIXTURE_DATA_2021_06_01 = "PVPC_CURV_DD_2021_06_01.json"
_FIXTURE_ESIOS_PVPC_2021_10_30 = "PRICES_ESIOS_1001_2021_10_30.json"
_FIXTURE_ESIOS_PVPC_2021_10_31 = "PRICES_ESIOS_1001_2021_10_31.json"
_FIXTURE_ESIOS_PVPC_2021_06_01 = "PRICES_ESIOS_1001_2021_06_01.json"
_FIXTURE_ESIOS_PVPC_2024_03_09 = "PRICES_ESIOS_1001_2024_03_09.json"
_FIXTURE_ESIOS_INJECTION_2021_10_30 = "PRICES_ESIOS_1739_2021_10_30.json"
_FIXTURE_ESIOS_INJECTION_2021_10_31 = "PRICES_ESIOS_1739_2021_10_31.json"
_FIXTURE_ESIOS_INJECTION_2024_03_09 = "PRICES_ESIOS_1739_2024_03_09.json"
_FIXTURE_ESIOS_OMIE_2021_10_30 = "PRICES_ESIOS_10211_2021_10_30.json"
_FIXTURE_ESIOS_OMIE_2021_10_31 = "PRICES_ESIOS_10211_2021_10_31.json"
_FIXTURE_ESIOS_OMIE_2024_03_09 = "PRICES_ESIOS_10211_2024_03_09.json"
_FIXTURE_ESIOS_MAG_2024_03_09 = "PRICES_ESIOS_1900_2024_03_09.json"
_FIXTURE_ESIOS_ADJUSTMENT_2024_03_09 = "PRICES_ESIOS_2108_2024_03_09.json"

_DEFAULT_EMPTY_VALUE = {"message": "No values for specified archive"}
_DEFAULT_UNAUTH_MSG = "HTTP Token: Access denied (TEST)."


class MockAsyncSession:
    """Mock GET requests to esios API."""

    status: int = 200
    _counter: int = 0
    _raw_response = None

    def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    def __await__(self):
        yield
        return self

    async def close(self, *_args):
        pass

    def __init__(self, status=200, exc=None):
        """Set up desired mock response"""
        self._raw_response = _DEFAULT_EMPTY_VALUE
        self.status = status
        self.exc = exc

        self.responses_public = {
            date(2022, 3, 27): load_fixture(_FIXTURE_DATA_2022_03_27),
            date(2021, 10, 30): load_fixture(_FIXTURE_DATA_2021_10_30),
            date(2021, 10, 31): load_fixture(_FIXTURE_DATA_2021_10_31),
            date(2021, 6, 1): load_fixture(_FIXTURE_DATA_2021_06_01),
        }
        self.responses_token = {
            ESIOS_PVPC: {
                date(2021, 10, 30): load_fixture(_FIXTURE_ESIOS_PVPC_2021_10_30),
                date(2021, 10, 31): load_fixture(_FIXTURE_ESIOS_PVPC_2021_10_31),
                date(2021, 6, 1): load_fixture(_FIXTURE_ESIOS_PVPC_2021_06_01),
                date(2024, 3, 9): load_fixture(_FIXTURE_ESIOS_PVPC_2024_03_09),
            },
            ESIOS_INJECTION: {
                date(2021, 10, 30): load_fixture(_FIXTURE_ESIOS_INJECTION_2021_10_30),
                date(2021, 10, 31): load_fixture(_FIXTURE_ESIOS_INJECTION_2021_10_31),
                date(2024, 3, 9): load_fixture(_FIXTURE_ESIOS_INJECTION_2024_03_09),
            },
            ESIOS_MAG: {
                date(2024, 3, 9): load_fixture(_FIXTURE_ESIOS_MAG_2024_03_09),
            },
            ESIOS_OMIE: {
                date(2021, 10, 30): load_fixture(_FIXTURE_ESIOS_OMIE_2021_10_30),
                date(2021, 10, 31): load_fixture(_FIXTURE_ESIOS_OMIE_2021_10_31),
                date(2024, 3, 9): load_fixture(_FIXTURE_ESIOS_OMIE_2024_03_09),
            },
            ESIOS_MARKET_ADJUSTMENT: {
                date(2024, 3, 9): load_fixture(_FIXTURE_ESIOS_ADJUSTMENT_2024_03_09)
            },
        }

    async def json(self, *_args, **_kwargs):
        """Dumb await."""
        return self._raw_response

    async def get(self, url: str, *_args, **_kwargs):
        """Dumb await."""
        self._counter += 1
        if self.exc:
            raise self.exc

        prefix_public = "https://api.esios.ree.es/archives/"
        prefix_token = "https://api.esios.ree.es/indicators/"
        key = datetime.fromisoformat(url.split("=")[-1]).date()
        if url.startswith(prefix_token):
            indicator = url.removeprefix(prefix_token).split("?")[0]
            self._raw_response = self.responses_token.get(indicator, {}).get(
                key, _DEFAULT_UNAUTH_MSG
            )
        elif url.startswith(prefix_public) and key in self.responses_public:
            self._raw_response = self.responses_public[key]
        else:
            self._raw_response = _DEFAULT_EMPTY_VALUE
        return self

    @property
    def call_count(self) -> int:
        """Return call counter."""
        return self._counter


def load_fixture(filename: str):
    """Load stored example for esios API response."""
    return json.loads((TEST_EXAMPLES_PATH / filename).read_text())


async def run_h_step(
    mock_session: MockAsyncSession,
    pvpc_data: PVPCData,
    api_data: EsiosApiData | None,
    start: datetime,
    should_fail: bool = False,
) -> tuple[datetime, EsiosApiData]:
    current_prices = api_data.sensors[KEY_PVPC] if api_data else {}
    if current_prices:
        logging.debug(
            "[Calls=%d]-> start=%s --> %s -> %s (%d prices)",
            mock_session.call_count,
            start,
            next(iter(current_prices)).strftime("%Y-%m-%d %Hh"),
            list(current_prices)[-1].strftime("%Y-%m-%d %Hh"),
            len(current_prices),
        )
    api_data = await pvpc_data.async_update_all(api_data, start)
    state_ok = pvpc_data.process_state_and_attributes(api_data, KEY_PVPC, start)
    assert should_fail is not state_ok
    start += timedelta(hours=1)
    return start, api_data


def check_num_datapoints(
    api_data: EsiosApiData, sensor_keys: tuple[str, ...], expected: int
):
    for key in sensor_keys:
        num_points = len(api_data.sensors[key])
        assert num_points == expected, (key, expected, num_points)



"""Tests for aiopvpc."""

from datetime import datetime, timedelta

import pytest

from aiopvpc.const import REFERENCE_TZ
from aiopvpc.pvpc_tariff import get_current_and_next_tariff_periods


@pytest.mark.parametrize(
    "year, days_weekend_p3, extra_days_p3",
    (
        (2021, 104, 7),
        (2022, 105, 7),
        (2023, 105, 9),
        (2024, 104, 7),
        (2025, 104, 7),
    ),
)
def test_number_of_national_holidays(year, days_weekend_p3, extra_days_p3):
    """Calculate days with full P3 valley period."""
    holidays_p3 = weekend_days_p3 = 0
    day = datetime(year, 1, 1, 15, tzinfo=REFERENCE_TZ)
    while day.year == year:
        period, _, _ = get_current_and_next_tariff_periods(day, False)
        if period == "P3":
            if day.isoweekday() > 5:
                weekend_days_p3 += 1
            else:
                holidays_p3 += 1
        day += timedelta(days=1)
    assert weekend_days_p3 == days_weekend_p3
    assert holidays_p3 == extra_days_p3



"""ESIOS API handler for HomeAssistant. Hourly price attributes."""

import zoneinfo
from contextlib import suppress
from datetime import datetime
from typing import Any

from .const import EsiosApiData, KEY_ADJUSTMENT, KEY_INDEXED, KEY_INJECTION, KEY_PVPC


def _is_tomorrow_price(ts: datetime, ref: datetime) -> bool:
    return any(
        ts_comp > ts_tz_ref
        for ts_comp, ts_tz_ref in zip(ts.isocalendar(), ref.isocalendar())
    )


def _split_today_tomorrow_prices(
    current_prices: dict[datetime, float],
    utc_time: datetime,
    timezone: zoneinfo.ZoneInfo,
) -> tuple[dict[datetime, float], dict[datetime, float]]:
    local_time = utc_time.astimezone(timezone)
    today, tomorrow = {}, {}
    for ts_utc, price_h in current_prices.items():
        ts_local = ts_utc.astimezone(timezone)
        if _is_tomorrow_price(ts_local, local_time):
            tomorrow[ts_utc] = price_h
        else:
            today[ts_utc] = price_h
    return today, tomorrow


def _make_price_tag_attributes(
    prices: dict[datetime, float], timezone: zoneinfo.ZoneInfo, tomorrow: bool
) -> dict[str, Any]:
    prefix = "price_next_day_" if tomorrow else "price_"
    attributes = {}
    for ts_utc, price_h in prices.items():
        ts_local = ts_utc.astimezone(timezone)
        attr_key = f"{prefix}{ts_local.hour:02d}h"
        if attr_key in attributes:  # DST change with duplicated hour :)
            attr_key += "_d"
        attributes[attr_key] = price_h
    return attributes


def _make_price_stats_attributes(
    sensor_key: str,
    current_price: float,
    current_prices: dict[datetime, float],
    utc_time: datetime,
    timezone: zoneinfo.ZoneInfo,
) -> dict[str, Any]:
    attributes: dict[str, Any] = {}
    sign_is_best = 1 if sensor_key != KEY_INJECTION else -1
    prices_sorted = dict(
        sorted(current_prices.items(), key=lambda x: sign_is_best * x[1])
    )
    better_prices_ahead = [
        (ts, price)
        for ts, price in current_prices.items()
        if ts > utc_time and price * sign_is_best < current_price * sign_is_best
    ]
    if better_prices_ahead:
        next_better_ts, next_better_price = better_prices_ahead[0]
        delta_better = next_better_ts - utc_time
        attributes["next_better_price"] = next_better_price
        attributes["hours_to_better_price"] = int(delta_better.total_seconds()) // 3600
        attributes["num_better_prices_ahead"] = len(better_prices_ahead)

    with suppress(ValueError):
        attributes["price_position"] = (
            list(prices_sorted.values()).index(current_price) + 1
        )

    max_price = max(current_prices.values())
    min_price = min(current_prices.values())
    with suppress(ZeroDivisionError):
        attributes["price_ratio"] = round(
            (current_price - min_price) / (max_price - min_price), 2
        )

    attributes["max_price"] = max_price
    first_price_at = next(iter(prices_sorted)).astimezone(timezone).hour
    last_price_at = next(iter(reversed(prices_sorted))).astimezone(timezone).hour
    attributes["max_price_at"] = last_price_at if sign_is_best == 1 else first_price_at
    attributes["min_price"] = min_price
    attributes["min_price_at"] = first_price_at if sign_is_best == 1 else last_price_at
    attributes["next_best_at"] = [
        ts.astimezone(timezone).hour for ts in prices_sorted if ts >= utc_time
    ]
    return attributes


def make_price_sensor_attributes(
    sensor_key: str,
    current_prices: dict[datetime, float],
    utc_time: datetime,
    timezone: zoneinfo.ZoneInfo,
) -> dict[str, Any]:
    """Generate sensor attributes for hourly prices variables."""
    current_price = current_prices[utc_time]
    today, tomorrow = _split_today_tomorrow_prices(current_prices, utc_time, timezone)
    price_attrs = _make_price_stats_attributes(
        sensor_key, current_price, today, utc_time, timezone
    )
    price_tags = _make_price_tag_attributes(today, timezone, False)
    if tomorrow:
        tomorrow_prices = {
            f"{key} (next day)": value
            for key, value in _make_price_stats_attributes(
                sensor_key, current_price, tomorrow, utc_time, timezone
            ).items()
        }
        tomorrow_price_tags = _make_price_tag_attributes(tomorrow, timezone, True)
        price_attrs = {**price_attrs, **tomorrow_prices}
        price_tags = {**price_tags, **tomorrow_price_tags}
    return {**price_attrs, **price_tags}


def add_composed_price_sensors(data: EsiosApiData):
    """Calculate price sensors derived from multiple data series."""
    if (
        data.availability.get(KEY_PVPC, False)
        and data.availability.get(KEY_ADJUSTMENT, False)
        and (
            common_ts_prices := set(data.sensors[KEY_PVPC]).intersection(
                set(data.sensors[KEY_ADJUSTMENT])
            )
        )
    ):
        # generate 'indexed tariff' as: PRICE = PVPC - ADJUSTMENT
        pvpc = data.sensors[KEY_PVPC]
        adjustment = data.sensors[KEY_ADJUSTMENT]
        data.sensors[KEY_INDEXED] = {
            ts_hour: round(pvpc[ts_hour] - adjustment[ts_hour], 5)
            for ts_hour in sorted(common_ts_prices)
        }
        data.availability[KEY_INDEXED] = True



"""Simple aio library to download Spanish electricity hourly prices. Constants."""

import zoneinfo
from dataclasses import dataclass
from datetime import date, datetime
from typing import Literal

DATE_CHANGE_TO_20TD = date(2021, 6, 1)

# Tariffs as internal keys in esios API data
TARIFF_20TD_IDS = ["PCB", "CYM"]

# Tariff names used in HomeAssistant integration
TARIFFS = ["2.0TD", "2.0TD (Ceuta/Melilla)"]

TARIFF2ID = dict(zip(TARIFFS, TARIFF_20TD_IDS))

# Contracted power
DEFAULT_POWER_KW = 3.3

# Prices are given in 0 to 24h sets, adjusted to the main timezone in Spain
REFERENCE_TZ = zoneinfo.ZoneInfo("Europe/Madrid")
UTC_TZ = zoneinfo.ZoneInfo("UTC")

DEFAULT_TIMEOUT = 10
PRICE_PRECISION = 5

DataSource = Literal["esios_public", "esios"]
GEOZONES = ["Península", "Canarias", "Baleares", "Ceuta", "Melilla", "España"]
GEOZONE_ID2NAME: dict[int, str] = {
    3: "España",
    8741: "Península",
    8742: "Canarias",
    8743: "Baleares",
    8744: "Ceuta",
    8745: "Melilla",
}
URL_PUBLIC_PVPC_RESOURCE = (
    "https://api.esios.ree.es/archives/70/download_json"
    "?locale=es&date={day:%Y-%m-%d}"
)
URL_ESIOS_TOKEN_RESOURCE = (
    "https://api.esios.ree.es/indicators/{ind}"
    + "?start_date={day:%Y-%m-%d}T00:00&end_date={day:%Y-%m-%d}T23:59"
)
ATTRIBUTIONS: dict[DataSource, str] = {
    "esios_public": "Data retrieved from api.esios.ree.es by REE",
    "esios": "Data retrieved with API token from api.esios.ree.es by REE",
}

# api.esios.ree.es/indicators
ESIOS_PVPC = "1001"
ESIOS_INJECTION = "1739"
ESIOS_MAG = "1900"  # regargo GAS
ESIOS_OMIE = "10211"  # precio mayorista
ESIOS_MARKET_ADJUSTMENT = "2108"  # ajuste mercado Indexada VS PVPC

# unique ids for each series
KEY_PVPC = "PVPC"
KEY_INJECTION = "INJECTION"
KEY_MAG = "MAG"  # regargo GAS
KEY_OMIE = "OMIE"  # precio mayorista
KEY_ADJUSTMENT = "ADJUSTMENT"  # ajuste mercado
# composed sensors
KEY_INDEXED = "INDEXED"  # precio indexada (:= PVPC - ADJUSTMENT)

ALL_SENSORS = (KEY_PVPC, KEY_INJECTION, KEY_MAG, KEY_OMIE, KEY_ADJUSTMENT)
SENSOR_KEY_TO_DATAID = {
    KEY_PVPC: ESIOS_PVPC,
    KEY_INJECTION: ESIOS_INJECTION,
    KEY_MAG: ESIOS_MAG,
    KEY_OMIE: ESIOS_OMIE,
    KEY_ADJUSTMENT: ESIOS_MARKET_ADJUSTMENT,
}
SENSOR_KEY_TO_NAME = {
    KEY_PVPC: "PVPC T. 2.0TD",
    KEY_INJECTION: "Precio de la energía excedentaria",
    KEY_MAG: "2.0TD Excedente o déficit ajuste liquidación",
    KEY_OMIE: "Precio medio horario final suma",
    KEY_ADJUSTMENT: "Ajuste de mercado a plazo",
    KEY_INDEXED: "Precio tarifa Indexada",
}
# API indicator dependencies for each price sensor
SENSOR_KEY_TO_API_SERIES = {
    KEY_PVPC: [KEY_PVPC],
    KEY_INJECTION: [KEY_INJECTION],
    KEY_MAG: [KEY_MAG],
    KEY_OMIE: [KEY_OMIE],
    KEY_ADJUSTMENT: [KEY_ADJUSTMENT],
    KEY_INDEXED: [KEY_PVPC, KEY_ADJUSTMENT],
}


@dataclass
class EsiosResponse:
    """Data schema for parsed data coming from ESIOS API."""

    name: str
    data_id: str
    last_update: datetime
    unit: str
    series: dict[str, dict[datetime, float]]


@dataclass
class EsiosApiData:
    """Data schema to store multiple series from ESIOS API."""

    last_update: datetime
    data_source: str
    sensors: dict[str, dict[datetime, float]]
    availability: dict[str, bool]



"""ESIOS API handler for HomeAssistant. PVPC tariff periods."""

from __future__ import annotations

from datetime import date, datetime, timedelta

_HOURS_P2 = (8, 9, 14, 15, 16, 17, 22, 23)
_HOURS_P2_CYM = (8, 9, 10, 15, 16, 17, 18, 23)
# TODO review 'festivos nacionales no sustituibles de fecha fija', + 6/1
# obtained from `holidays` library,
# - with weekend days disabled (already full P3)
# - no 'translated' holidays
# - no 'Jueves Santo' as special day
_NATIONAL_EXTRA_HOLIDAYS_FOR_P3_PERIOD = {
    2021: {
        date(2021, 1, 1): "(viernes), Año nuevo",
        date(2021, 1, 6): "(miércoles), Epifanía del Señor",
        # date(2021, 4, 1): "(jueves), Jueves Santo",
        date(2021, 4, 2): "(viernes), Viernes Santo",
        # date(2021, 5, 1): "(sábado), Día del Trabajador",
        # date(2021, 8, 15): "(domingo), Asunción de la Virgen",
        date(2021, 10, 12): "(martes), Día de la Hispanidad",
        date(2021, 11, 1): "(lunes), Todos los Santos",
        date(2021, 12, 6): "(lunes), Día de la Constitución Española",
        date(2021, 12, 8): "(miércoles), La Inmaculada Concepción",
        # date(2021, 12, 25): "(sábado), Navidad",
    },
    2022: {
        # date(2022, 1, 1): "(sábado), Año nuevo",
        date(2022, 1, 6): "(jueves), Epifanía del Señor",
        date(2022, 4, 15): "(viernes), Viernes Santo",
        date(2022, 8, 15): "(lunes), Asunción de la Virgen",
        date(2022, 10, 12): "(miércoles), Día de la Hispanidad",
        date(2022, 11, 1): "(martes), Todos los Santos",
        date(2022, 12, 6): "(martes), Día de la Constitución Española",
        date(2022, 12, 8): "(jueves), La Inmaculada Concepción",
        # date(2022, 12, 26): "(lunes), Navidad (Trasladado)",
    },
    2023: {
        # date(2023, 1, 1): "(domingo), Año nuevo",
        date(2023, 1, 6): "(viernes), Epifanía del Señor",
        # date(2023, 4, 6): "(jueves), Jueves Santo",
        date(2023, 4, 7): "(viernes), Viernes Santo",
        date(2023, 5, 1): "(lunes), Día del Trabajador",
        date(2023, 8, 15): "(martes), Asunción de la Virgen",
        date(2023, 10, 12): "(jueves), Día de la Hispanidad",
        date(2023, 11, 1): "(miércoles), Todos los Santos",
        date(2023, 12, 6): "(miércoles), Día de la Constitución Española",
        date(2023, 12, 8): "(viernes), La Inmaculada Concepción",
        date(2023, 12, 25): "(lunes), Navidad",
    },
    2024: {
        date(2024, 1, 1): "(lunes), Año nuevo",
        # date(2024, 1, 6): "(sábado), Epifanía del Señor",
        # date(2024, 3, 28): "(jueves), Jueves Santo",
        date(2024, 3, 29): "(viernes), Viernes Santo",
        date(2024, 5, 1): "(miércoles), Día del Trabajador",
        date(2024, 8, 15): "(jueves), Asunción de la Virgen",
        # date(2024, 10, 12): "(sábado), Día de la Hispanidad",
        date(2024, 11, 1): "(viernes), Todos los Santos",
        date(2024, 12, 6): "(viernes), Día de la Constitución Española",
        # date(2024, 12, 8): "(domingo), La Inmaculada Concepción",
        date(2024, 12, 25): "(miércoles), Navidad",
    },
    2025: {
        date(2025, 1, 1): "(miércoles), Año nuevo",
        date(2025, 1, 6): "(lunes), Epifanía del Señor",
        # date(2025, 4, 17): "(jueves), Jueves Santo",
        date(2025, 4, 18): "(viernes), Viernes Santo",
        date(2025, 5, 1): "(jueves), Día del Trabajador",
        date(2025, 8, 15): "(viernes), Asunción de la Virgen",
        # date(2025, 10, 12): "(domingo), Día de la Hispanidad",
        # date(2025, 11, 1): "(sábado), Todos los Santos",
        # date(2025, 12, 6): "(sábado), Día de la Constitución Española",
        date(2025, 12, 8): "(lunes), La Inmaculada Concepción",
        date(2025, 12, 25): "(jueves), Navidad",
    },
}


def _tariff_period_key(local_ts: datetime, zone_ceuta_melilla: bool) -> str:
    """Return period key (P1/P2/P3) for current hour."""
    day = local_ts.date()
    national_holiday = day in _NATIONAL_EXTRA_HOLIDAYS_FOR_P3_PERIOD[day.year]
    if national_holiday or day.isoweekday() >= 6 or local_ts.hour < 8:
        return "P3"
    if zone_ceuta_melilla and local_ts.hour in _HOURS_P2_CYM:
        return "P2"
    if not zone_ceuta_melilla and local_ts.hour in _HOURS_P2:
        return "P2"
    return "P1"


def get_current_and_next_tariff_periods(
    local_ts: datetime, zone_ceuta_melilla: bool
) -> tuple[str, str, timedelta]:
    """Get tariff periods for PVPC 2.0TD."""
    current_period = _tariff_period_key(local_ts, zone_ceuta_melilla)
    delta = timedelta(hours=1)
    while (
        next_period := _tariff_period_key(local_ts + delta, zone_ceuta_melilla)
    ) == current_period:
        delta += timedelta(hours=1)
    return current_period, next_period, delta



"""Simple aio library to download Spanish electricity hourly prices."""

from .const import DEFAULT_POWER_KW, EsiosApiData, TARIFFS
from .ha_helpers import get_enabled_sensor_keys
from .pvpc_data import BadApiTokenAuthError, PVPCData

__all__ = (
    "BadApiTokenAuthError",
    "EsiosApiData",
    "DEFAULT_POWER_KW",
    "PVPCData",
    "TARIFFS",
    "get_enabled_sensor_keys",
)



"""Tests for aiopvpc."""

import os
from datetime import datetime
from typing import cast
from zoneinfo import ZoneInfo

import pytest
from aiohttp import ClientSession
from dotenv import load_dotenv

from aiopvpc import PVPCData
from aiopvpc.const import ALL_SENSORS, DataSource, KEY_PVPC, REFERENCE_TZ
from tests.conftest import TZ_TEST

load_dotenv()


async def _get_real_data(
    timezone: ZoneInfo, data_source: str, indicators: tuple[str, ...], ts: datetime
):
    async with ClientSession() as session:
        pvpc_data = PVPCData(
            session=session,
            tariff="2.0TD",
            local_timezone=timezone,
            api_token=os.getenv("ESIOS_TOKEN") if data_source == "esios" else None,
            data_source=cast(DataSource, data_source),
            sensor_keys=indicators,
        )
        return await pvpc_data.async_update_all(None, ts)


@pytest.mark.real_api_call
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "data_source, timezone, num_sensors",
    (
        ("esios", REFERENCE_TZ, 6),
        ("esios", TZ_TEST, 6),
        ("esios_public", REFERENCE_TZ, 1),
        ("esios_public", TZ_TEST, 1),
    ),
)
async def test_real_download_today_async(data_source, timezone, num_sensors):
    sensor_keys = ALL_SENSORS if data_source == "esios" else (KEY_PVPC,)
    api_data = await _get_real_data(
        timezone, data_source, sensor_keys, datetime.utcnow()
    )
    assert 22 < len(api_data.sensors[KEY_PVPC]) < 49
    assert len(api_data.sensors) == num_sensors


if __name__ == "__main__":
    import asyncio
    from dataclasses import asdict
    from pprint import pprint

    # timestamp = datetime(2021, 10, 30, 21)
    timestamp = datetime.utcnow()
    api_data = asyncio.run(
        _get_real_data(REFERENCE_TZ, "esios", ALL_SENSORS, timestamp)
    )
    pprint(asdict(api_data))



"""
Simple aio library to download Spanish electricity hourly prices.

* URL for JSON daily files
* Parser for the contents of the JSON files
"""

from datetime import datetime, timedelta
from itertools import groupby
from operator import itemgetter
from typing import Any

from aiopvpc.const import (
    DataSource,
    EsiosResponse,
    GEOZONE_ID2NAME,
    GEOZONES,
    KEY_PVPC,
    PRICE_PRECISION,
    REFERENCE_TZ,
    SENSOR_KEY_TO_DATAID,
    TARIFF2ID,
    TARIFFS,
    URL_ESIOS_TOKEN_RESOURCE,
    URL_PUBLIC_PVPC_RESOURCE,
    UTC_TZ,
    zoneinfo,
)


def _timezone_offset(tz: zoneinfo.ZoneInfo = REFERENCE_TZ) -> timedelta:
    ref_ts = datetime(2021, 1, 1, tzinfo=REFERENCE_TZ).astimezone(UTC_TZ)
    loc_ts = datetime(2021, 1, 1, tzinfo=tz).astimezone(UTC_TZ)
    return loc_ts - ref_ts


def extract_prices_from_esios_public(
    data: dict[str, Any], key: str, tz: zoneinfo.ZoneInfo = REFERENCE_TZ
) -> EsiosResponse:
    """Parse the contents of a daily PVPC json file."""
    ts_init = datetime(
        *datetime.strptime(data["PVPC"][0]["Dia"], "%d/%m/%Y").timetuple()[:3],
        tzinfo=tz,
    ).astimezone(UTC_TZ)

    def _parse_tariff_val(value, prec=PRICE_PRECISION) -> float:
        return round(float(value.replace(",", ".")) / 1000.0, prec)

    pvpc_prices = {
        ts_init + timedelta(hours=i): _parse_tariff_val(values_hour[key])
        for i, values_hour in enumerate(data["PVPC"])
    }

    return EsiosResponse(
        name="PVPC ESIOS",
        data_id="legacy",
        last_update=datetime.utcnow().replace(microsecond=0, tzinfo=UTC_TZ),
        unit="€/kWh",
        series={KEY_PVPC: pvpc_prices},
    )


def extract_prices_from_esios_token(
    data: dict[str, Any],
    sensor_key: str,
    geo_zone: str,
    tz: zoneinfo.ZoneInfo = REFERENCE_TZ,
) -> EsiosResponse:
    """Parse the contents of an 'indicator' json file from ESIOS API."""
    offset_timezone = _timezone_offset(tz)
    indicator_data = data.pop("indicator")
    unit = "•".join(mag["name"] for mag in indicator_data["magnitud"])
    unit_tiempo = "•".join(mag["name"] for mag in indicator_data["tiempo"])
    unit += f"/{unit_tiempo}"
    ts_update = datetime.utcnow().replace(microsecond=0, tzinfo=UTC_TZ)

    def _parse_dt(ts: str) -> datetime:
        return datetime.fromisoformat(ts).astimezone(UTC_TZ) + offset_timezone

    def _value_unit_conversion(value: float) -> float:
        # from €/MWh to €/kWh
        return round(float(value) / 1000.0, 5)

    value_gen = groupby(
        sorted(indicator_data["values"], key=itemgetter("geo_id")),
        itemgetter("geo_id"),
    )
    parsed_data = {
        GEOZONE_ID2NAME[key]: {
            _parse_dt(item["datetime"]): _value_unit_conversion(item["value"])
            for item in list(group)
        }
        for key, group in value_gen
    }
    if geo_zone in parsed_data:
        geo_data = parsed_data[geo_zone]
    elif "Península" in parsed_data:
        geo_data = parsed_data["Península"]
    else:
        geo_data = parsed_data["España"]

    return EsiosResponse(
        name=indicator_data["name"],
        data_id=str(indicator_data["id"]),
        last_update=ts_update,
        unit=unit,
        series={sensor_key: geo_data},
    )


def extract_esios_data(
    data: dict[str, Any],
    url: str,
    sensor_key: str,
    tariff: str,
    tz: zoneinfo.ZoneInfo = REFERENCE_TZ,
) -> EsiosResponse:
    """Parse the contents of a daily PVPC json file."""
    if url.startswith("https://api.esios.ree.es/archives"):
        return extract_prices_from_esios_public(data, TARIFF2ID[tariff], tz)

    if url.startswith("https://api.esios.ree.es/indicators"):
        # TODO adapt better to geozones
        if tariff == TARIFFS[0] and tz != REFERENCE_TZ:
            geo_zone = GEOZONES[1]
        elif tariff == TARIFFS[0]:
            geo_zone = GEOZONES[0]
        else:
            geo_zone = GEOZONES[3]

        return extract_prices_from_esios_token(data, sensor_key, geo_zone, tz)
    raise NotImplementedError(f"Data source not known: {url} >{data}")


def get_daily_urls_to_download(
    source: DataSource,
    sensor_keys: set[str],
    now_local_ref: datetime,
    next_day_local_ref: datetime,
) -> tuple[list[str], list[str]]:
    """Make URLs for ESIOS price series."""
    if source == "esios_public":
        assert sensor_keys == {KEY_PVPC}
        return (
            [URL_PUBLIC_PVPC_RESOURCE.format(day=now_local_ref.date())],
            [URL_PUBLIC_PVPC_RESOURCE.format(day=next_day_local_ref.date())],
        )

    assert source == "esios"
    today = [
        URL_ESIOS_TOKEN_RESOURCE.format(
            ind=SENSOR_KEY_TO_DATAID[key], day=now_local_ref.date()
        )
        for key in sensor_keys
    ]
    tomorrow = [
        URL_ESIOS_TOKEN_RESOURCE.format(
            ind=SENSOR_KEY_TO_DATAID[key], day=next_day_local_ref.date()
        )
        for key in sensor_keys
    ]
    return today, tomorrow



"""Tests for aiopvpc."""

from __future__ import annotations

import logging
from datetime import datetime

import pytest

from aiopvpc.const import (
    ALL_SENSORS,
    KEY_ADJUSTMENT,
    KEY_INJECTION,
    KEY_MAG,
    KEY_OMIE,
    KEY_PVPC,
    TARIFFS,
    UTC_TZ,
)
from aiopvpc.ha_helpers import get_enabled_sensor_keys, make_sensor_unique_id
from aiopvpc.pvpc_data import BadApiTokenAuthError, PVPCData
from tests.conftest import check_num_datapoints, MockAsyncSession, run_h_step


def test_sensor_unique_ids():
    all_sensor_keys = get_enabled_sensor_keys(
        using_private_api=False, disabled_sensor_ids=[]
    )
    assert all_sensor_keys == {KEY_PVPC}

    all_sensor_keys = get_enabled_sensor_keys(
        using_private_api=True, disabled_sensor_ids=[]
    )
    assert sorted(all_sensor_keys) == sorted(ALL_SENSORS)
    counter_combs = 0
    unique_ids = set()
    for tariff in TARIFFS:
        for key in all_sensor_keys:
            unique_ids.add(
                make_sensor_unique_id(config_entry_id=tariff, sensor_key=key)
            )
            counter_combs += 1
    assert counter_combs == len(unique_ids)

    assert not get_enabled_sensor_keys(
        using_private_api=True, disabled_sensor_ids=list(unique_ids)
    )


@pytest.mark.asyncio
async def test_disable_sensors():
    start = datetime(2024, 3, 9, 19, tzinfo=UTC_TZ)
    mock_session = MockAsyncSession()
    sensor_keys = ALL_SENSORS
    assert len(sensor_keys) == 5
    pvpc_data = PVPCData(
        session=mock_session,
        tariff="2.0TD",
        data_source="esios",
        api_token="test-token",
        sensor_keys=sensor_keys,
    )

    api_data = None
    start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
    assert mock_session.call_count == 10
    check_num_datapoints(api_data, sensor_keys, 24)

    pvpc_data.update_active_sensors(KEY_PVPC, enabled=False)
    pvpc_data.update_active_sensors(KEY_OMIE, enabled=False)
    pvpc_data.update_active_sensors(KEY_MAG, enabled=False)
    pvpc_data.update_active_sensors(KEY_MAG, enabled=False)
    pvpc_data.update_active_sensors(KEY_ADJUSTMENT, enabled=False)

    start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
    assert mock_session.call_count == 11
    check_num_datapoints(api_data, sensor_keys, 24)
    logging.error(api_data.sensors.keys())

    pvpc_data.update_active_sensors(KEY_INJECTION, enabled=False)
    start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
    assert mock_session.call_count == 11
    check_num_datapoints(api_data, sensor_keys, 24)

    start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
    assert mock_session.call_count == 11
    check_num_datapoints(api_data, sensor_keys, 24)

    start, api_data = await run_h_step(
        mock_session, pvpc_data, api_data, start, should_fail=True
    )
    assert mock_session.call_count == 11
    # check_num_datapoints(api_data, sensor_keys, 0)

    pvpc_data.update_active_sensors(KEY_INJECTION, enabled=True)
    pvpc_data.update_active_sensors(KEY_PVPC, enabled=True)
    start, api_data = await run_h_step(
        mock_session, pvpc_data, api_data, start, should_fail=True
    )
    assert mock_session.call_count == 13

    pvpc_data.update_active_sensors(KEY_INJECTION, enabled=True)
    pvpc_data.update_active_sensors(KEY_PVPC, enabled=True)
    pvpc_data.update_active_sensors(KEY_MAG, enabled=True)
    pvpc_data.update_active_sensors(KEY_OMIE, enabled=True)
    await run_h_step(mock_session, pvpc_data, api_data, start, should_fail=True)
    assert mock_session.call_count == 17


@pytest.mark.asyncio
async def test_check_api_token():
    start = datetime(2024, 3, 9, 19, tzinfo=UTC_TZ)
    mock_session = MockAsyncSession(status=401)
    pvpc_data = PVPCData(session=mock_session)
    token_ok = await pvpc_data.check_api_token(start, "bad_token")
    assert not token_ok
    assert mock_session.call_count == 1

    with pytest.raises(BadApiTokenAuthError):
        await pvpc_data.async_update_all(None, start)
    assert mock_session.call_count == 2

    mock_session_ok = MockAsyncSession()
    pvpc_data_ok = PVPCData(
        session=mock_session_ok,
        data_source="esios",
        api_token="good_token",
    )
    token_ok = await pvpc_data_ok.check_api_token(start)
    assert token_ok
    assert mock_session_ok.call_count == 1



"""
Simple aio library to download Spanish electricity hourly prices.

Externalization of download and parsing logic for the `pvpc_hourly_pricing`
HomeAssistant integration,
https://www.home-assistant.io/integrations/pvpc_hourly_pricing/
"""

from __future__ import annotations

import asyncio
import logging
from collections import deque
from datetime import datetime, timedelta
from random import random
from typing import Any

import aiohttp
import async_timeout

from aiopvpc.const import (
    ALL_SENSORS,
    ATTRIBUTIONS,
    DataSource,
    DEFAULT_POWER_KW,
    DEFAULT_TIMEOUT,
    EsiosApiData,
    EsiosResponse,
    KEY_PVPC,
    REFERENCE_TZ,
    SENSOR_KEY_TO_API_SERIES,
    SENSOR_KEY_TO_DATAID,
    TARIFFS,
    UTC_TZ,
    zoneinfo,
)
from aiopvpc.parser import extract_esios_data, get_daily_urls_to_download
from aiopvpc.prices import add_composed_price_sensors, make_price_sensor_attributes
from aiopvpc.pvpc_tariff import get_current_and_next_tariff_periods
from aiopvpc.utils import ensure_utc_time

_LOGGER = logging.getLogger(__name__)

# TODO REMOVE THIS USER-AGENT LOGIC
# 🙈😱 Use randomized standard User-Agent info to avoid server banning 😖🤷
_STANDARD_USER_AGENTS = [
    (
        "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:47.0) "
        "Gecko/20100101 Firefox/47.3"
    ),
    (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X x.y; rv:42.0) "
        "Gecko/20100101 Firefox/43.4"
    ),
    (
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 "
        "(KHTML, like Gecko) Chrome/77.0.3865.90 Safari/537.36"
    ),
    (
        "Mozilla/5.0 (iPhone; CPU iPhone OS 11_3_1 like Mac OS X) "
        "AppleWebKit/603.1.30 (KHTML, like Gecko)"
    ),
    "Version/10.0 Mobile/14E304 Safari/602.1",
]


class BadApiTokenAuthError(Exception):
    """Exception to signal HA that ESIOS API token is invalid (401 status)."""

    pass  # noqa PIE790


class PVPCData:
    """
    Data handler for PVPC hourly prices.

    * Async download of prices for each day
    * Generate state attributes for HA integration.

    - Prices are returned in a `dict[datetime, float]`,
    with timestamps in UTC and prices in €/kWh.
    """

    def __init__(
        self,
        *,
        session: aiohttp.ClientSession,
        tariff: str = TARIFFS[0],
        local_timezone: str | zoneinfo.ZoneInfo = REFERENCE_TZ,
        power: float = DEFAULT_POWER_KW,
        power_valley: float = DEFAULT_POWER_KW,
        timeout: float = DEFAULT_TIMEOUT,
        data_source: DataSource = "esios_public",
        api_token: str | None = None,
        sensor_keys: tuple[str, ...] = (KEY_PVPC,),
    ) -> None:
        """Set up API access."""
        self.states: dict[str, float | None] = {}
        self.sensor_attributes: dict[str, dict[str, Any]] = {}
        self._sensor_keys: set[str] = {key for key in sensor_keys if key in ALL_SENSORS}

        self._timeout = timeout
        self._session = session
        self._data_source = data_source
        self._api_token = api_token
        if self._api_token is not None:
            self._data_source = "esios"
        assert (data_source != "esios") or self._api_token is not None, data_source
        self._user_agents = deque(sorted(_STANDARD_USER_AGENTS, key=lambda _: random()))

        self._local_timezone = zoneinfo.ZoneInfo(str(local_timezone))
        assert tariff in TARIFFS
        self.tariff = tariff

        self._power = power
        self._power_valley = power_valley

    @property
    def using_private_api(self) -> bool:
        """Check if an API token is available and data-source is ESIOS."""
        return self._api_token is not None and self._data_source == "esios"

    async def _api_get_data(self, sensor_key: str, url: str) -> EsiosResponse | None:
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Host": "api.esios.ree.es",
            "User-Agent": self._user_agents[0],
        }
        if self.using_private_api:
            assert self._api_token is not None
            headers["x-api-key"] = self._api_token
            headers["Authorization"] = f"Token token={self._api_token}"

        assert self._session is not None
        resp = await self._session.get(url, headers=headers)
        if resp.status < 400:
            data = await resp.json()
            return extract_esios_data(
                data, url, sensor_key, self.tariff, tz=self._local_timezone
            )
        elif resp.status in (401, 403) and self._data_source == "esios":
            _LOGGER.warning(
                "[%s] Unauthorized error with '%s': %s",
                sensor_key,
                self._data_source,
                url,
            )
            raise BadApiTokenAuthError(
                f"[{sensor_key}] Unauthorized access with API token '{self._api_token}'"
            )
        elif resp.status == 403:  # pragma: no cover
            _LOGGER.warning(
                "[%s] Forbidden error with '%s': %s", sensor_key, self._data_source, url
            )
            # loop user-agent and data-source
            self._user_agents.rotate()
        else:
            _LOGGER.error(
                "[%s] Unknown error [%d] with '%s': %s",
                sensor_key,
                resp.status,
                self._data_source,
                url,
            )
        return None

    async def _download_daily_data(
        self, sensor_key: str, url: str
    ) -> EsiosResponse | None:
        """
        PVPC data extractor.

        Make GET request to 'api.esios.ree.es' and extract hourly prices.

        Prices are referenced with datetimes in UTC.
        """
        try:
            async with async_timeout.timeout(self._timeout):
                return await self._api_get_data(sensor_key, url)
        except (AttributeError, KeyError) as exc:
            _LOGGER.debug("[%s] Bad try on getting prices (%s)", sensor_key, exc)
        except asyncio.TimeoutError:
            _LOGGER.warning(
                "[%s] Timeout error requesting data from '%s'", sensor_key, url
            )
        except aiohttp.ClientError as exc:
            _LOGGER.warning("[%s] Client error in '%s' -> %s", sensor_key, url, exc)
        except BadApiTokenAuthError:
            raise
        return None

    async def check_api_token(
        self, now: datetime, api_token: str | None = None
    ) -> bool:
        """Check if ESIOS API token is valid."""
        local_ref_now = ensure_utc_time(now).astimezone(REFERENCE_TZ)
        if api_token is not None:
            self._api_token = api_token
        self._data_source = "esios"
        today, _ = get_daily_urls_to_download(
            self._data_source,
            {KEY_PVPC},
            local_ref_now,
            local_ref_now,
        )
        try:
            prices = await self._download_daily_data(KEY_PVPC, today[0])
        except BadApiTokenAuthError:
            return False
        return prices is not None

    def update_active_sensors(self, data_id: str, enabled: bool):
        """Update enabled API indicators to download."""
        assert data_id in ALL_SENSORS
        if enabled:
            self._sensor_keys.add(data_id)
        elif data_id in self._sensor_keys:
            self._sensor_keys.remove(data_id)

    async def async_update_all(
        self, current_data: EsiosApiData | None, now: datetime
    ) -> EsiosApiData:
        """
        Update all prices from the ESIOS API.

        Input `now: datetime` is assumed tz-aware in UTC.
        If not, it is converted to UTC from the original timezone,
        or set as UTC-time if it is a naive datetime.
        """
        utc_now = ensure_utc_time(now)
        local_ref_now = utc_now.astimezone(REFERENCE_TZ)
        next_day = local_ref_now + timedelta(days=1)

        if current_data is None:
            current_data = EsiosApiData(
                sensors={},
                availability={},
                data_source=self._data_source,
                last_update=utc_now,
            )

        api_sensors = {
            api_sensor_key
            for sensor_key in self._sensor_keys
            for api_sensor_key in SENSOR_KEY_TO_API_SERIES[sensor_key]
        }
        urls_now, urls_next = get_daily_urls_to_download(
            self._data_source,
            api_sensors,
            local_ref_now,
            next_day,
        )
        updated = False
        tasks = []
        for url_now, url_next, sensor_key in zip(urls_now, urls_next, api_sensors):
            if sensor_key not in current_data.sensors:
                current_data.sensors[sensor_key] = {}

            tasks.append(
                self._update_prices_series(
                    sensor_key,
                    current_data.sensors[sensor_key],
                    url_now,
                    url_next,
                    local_ref_now,
                )
            )

        results = await asyncio.gather(*tasks)
        for new_data, sensor_key in zip(results, api_sensors):
            if new_data:
                updated = True
                current_data.sensors[sensor_key] = new_data
                current_data.availability[sensor_key] = True

        if updated:
            current_data.data_source = self._data_source
            current_data.last_update = utc_now

        add_composed_price_sensors(current_data)
        for sensor_key in current_data.sensors:
            self.process_state_and_attributes(current_data, sensor_key, now)
        return current_data

    async def _update_prices_series(
        self,
        sensor_key: str,
        current_prices: dict[datetime, float],
        url_now: str,
        url_next: str,
        local_ref_now: datetime,
    ) -> dict[datetime, float] | None:
        current_num_prices = len(current_prices)
        if local_ref_now.hour >= 20 and current_num_prices > 30:
            # already have today+tomorrow prices, avoid requests
            _LOGGER.debug(
                "[%s] Evening download avoided, now with %d prices from %s UTC",
                sensor_key,
                current_num_prices,
                next(iter(current_prices)).strftime("%Y-%m-%d %Hh"),
            )
            return None
        elif (
            local_ref_now.hour < 20
            and current_num_prices > 20
            and (
                list(current_prices)[-12].astimezone(REFERENCE_TZ).date()
                == local_ref_now.date()
            )
        ):
            # already have today prices, avoid request
            _LOGGER.debug(
                "[%s] Download avoided, now with %d prices up to %s UTC",
                sensor_key,
                current_num_prices,
                list(current_prices)[-1].strftime("%Y-%m-%d %Hh"),
            )
            return None

        if current_num_prices and (
            next(iter(current_prices)).astimezone(REFERENCE_TZ).date()
            == local_ref_now.date()
        ):
            # avoid download of today prices
            _LOGGER.debug(
                "[%s] Avoided: %s, with %d prices -> last: %s, download-day: %s",
                sensor_key,
                local_ref_now,
                current_num_prices,
                next(iter(current_prices)).astimezone(REFERENCE_TZ).date(),
                local_ref_now.date(),
            )
        else:
            # make API call to download today prices
            prices_response = await self._download_daily_data(sensor_key, url_now)
            if prices_response is None or not prices_response.series.get(sensor_key):
                return current_prices
            prices = prices_response.series[sensor_key]
            current_prices.update(prices)

        # At evening, it is possible to retrieve next day prices
        if local_ref_now.hour >= 20:
            prices_fut_response = await self._download_daily_data(sensor_key, url_next)
            if prices_fut_response:
                prices_fut = prices_fut_response.series[sensor_key]
                current_prices.update(prices_fut)

        _LOGGER.debug(
            "[%s] Download done, now with %d prices from %s UTC",
            sensor_key,
            len(current_prices),
            next(iter(current_prices)).strftime("%Y-%m-%d %Hh"),
        )

        return current_prices

    @property
    def attribution(self) -> str:
        """Return data-source attribution string."""
        return ATTRIBUTIONS[self._data_source]

    def process_state_and_attributes(
        self, current_data: EsiosApiData, sensor_key: str, utc_now: datetime
    ) -> bool:
        """
        Generate the current state and sensor attributes.

        The data source provides prices in 0 to 24h sets, with correspondence
        with the main timezone in Spain. They are stored with UTC datetimes.

        Input `now: datetime` is assumed tz-aware in UTC.
        If not, it is converted to UTC from the original timezone,
        or set as UTC-time if it is a naive datetime.
        """
        attributes: dict[str, Any] = {
            "sensor_id": sensor_key,
            "data_id": SENSOR_KEY_TO_DATAID.get(sensor_key, "composed"),
        }
        utc_time = ensure_utc_time(utc_now.replace(minute=0, second=0, microsecond=0))
        actual_time = utc_time.astimezone(self._local_timezone)
        current_prices = current_data.sensors.get(sensor_key, {})
        if len(current_prices) > 25 and actual_time.hour < 20:
            # there are 'today' and 'next day' prices, but 'today' has expired
            max_age = (
                utc_time.astimezone(REFERENCE_TZ).replace(hour=0).astimezone(UTC_TZ)
            )
            current_data.sensors[sensor_key] = {
                key_ts: price
                for key_ts, price in current_prices.items()
                if key_ts >= max_age
            }

        # set current price
        try:
            self.states[sensor_key] = current_data.sensors[sensor_key][utc_time]
            current_data.availability[sensor_key] = True
        except KeyError:
            self.states[sensor_key] = None
            current_data.availability[sensor_key] = False
            self.sensor_attributes[sensor_key] = attributes
            return False

        # generate price attributes
        price_attrs = make_price_sensor_attributes(
            sensor_key, current_data.sensors[sensor_key], utc_time, self._local_timezone
        )

        # generate PVPC 2.0TD sensor attributes
        if sensor_key == KEY_PVPC:
            local_time = utc_time.astimezone(self._local_timezone)
            (
                current_period,
                next_period,
                delta,
            ) = get_current_and_next_tariff_periods(
                local_time, zone_ceuta_melilla=self.tariff != TARIFFS[0]
            )
            attributes["tariff"] = self.tariff
            attributes["period"] = current_period
            power = self._power_valley if current_period == "P3" else self._power
            attributes["available_power"] = int(1000 * power)
            attributes["next_period"] = next_period
            attributes["hours_to_next_period"] = int(delta.total_seconds()) // 3600

        self.sensor_attributes[sensor_key] = {**attributes, **price_attrs}
        return True



"""ESIOS API handler for HomeAssistant. Utils."""

from datetime import datetime

from aiopvpc.const import UTC_TZ


def ensure_utc_time(ts: datetime) -> datetime:
    """Return tz-aware datetime in UTC from any datetime."""
    if ts.tzinfo is None:
        return datetime(*ts.timetuple()[:6], tzinfo=UTC_TZ)
    elif str(ts.tzinfo) != str(UTC_TZ):
        return ts.astimezone(UTC_TZ)
    return ts



"""Home Assistant helper methods."""

from aiopvpc.const import (
    ALL_SENSORS,
    KEY_ADJUSTMENT,
    KEY_INDEXED,
    KEY_INJECTION,
    KEY_MAG,
    KEY_OMIE,
    KEY_PVPC,
    TARIFFS,
)

_ha_uniqueid_to_sensor_key = {
    TARIFFS[0]: KEY_PVPC,
    TARIFFS[1]: KEY_PVPC,
    f"{TARIFFS[0]}_{KEY_INJECTION}": KEY_INJECTION,
    f"{TARIFFS[1]}_{KEY_INJECTION}": KEY_INJECTION,
    f"{TARIFFS[0]}_INYECTION": KEY_INJECTION,
    f"{TARIFFS[1]}_INYECTION": KEY_INJECTION,
    f"{TARIFFS[0]}_{KEY_MAG}": KEY_MAG,
    f"{TARIFFS[1]}_{KEY_MAG}": KEY_MAG,
    f"{TARIFFS[0]}_{KEY_OMIE}": KEY_OMIE,
    f"{TARIFFS[1]}_{KEY_OMIE}": KEY_OMIE,
    f"{TARIFFS[0]}_{KEY_ADJUSTMENT}": KEY_ADJUSTMENT,
    f"{TARIFFS[1]}_{KEY_ADJUSTMENT}": KEY_ADJUSTMENT,
    f"{TARIFFS[0]}_{KEY_INDEXED}": KEY_INDEXED,
    f"{TARIFFS[1]}_{KEY_INDEXED}": KEY_INDEXED,
}


def get_enabled_sensor_keys(
    using_private_api: bool, disabled_sensor_ids: list[str]
) -> set[str]:
    """(HA) Get enabled API indicators."""
    sensor_keys = set(ALL_SENSORS) if using_private_api else {KEY_PVPC}
    for unique_id in disabled_sensor_ids:
        disabled_ind = _ha_uniqueid_to_sensor_key[unique_id]
        if disabled_ind in sensor_keys:
            sensor_keys.remove(disabled_ind)

    return sensor_keys


def make_sensor_unique_id(config_entry_id: str, sensor_key: str) -> str:
    """(HA) Generate unique_id for each sensor kind and config entry."""
    assert sensor_key in ALL_SENSORS or sensor_key == KEY_INDEXED
    if sensor_key == KEY_PVPC:
        # for old compatibility
        return config_entry_id
    return f"{config_entry_id}_{sensor_key}"



"""Tests for aiopvpc."""

import logging
from asyncio import TimeoutError
from datetime import datetime, timedelta
from typing import cast

import pytest
from aiohttp import ClientError

from aiopvpc.const import (
    ALL_SENSORS,
    ATTRIBUTIONS,
    DataSource,
    KEY_INJECTION,
    KEY_OMIE,
    KEY_PVPC,
    REFERENCE_TZ,
    UTC_TZ,
)
from aiopvpc.pvpc_data import BadApiTokenAuthError, PVPCData
from tests.conftest import check_num_datapoints, MockAsyncSession, run_h_step, TZ_TEST


@pytest.mark.parametrize(
    "data_source, api_token, day_str, num_log_msgs, status, exception",
    (
        ("esios_public", None, "2032-10-26", 0, 200, None),
        ("esios_public", None, "2032-10-26", 1, 500, None),
        ("esios", "bad-token", "2032-10-26", 1, 403, None),
        ("esios", "bad-token", "2032-10-26", 1, 401, None),
        ("esios_public", None, "2032-10-26", 1, 200, TimeoutError),
        ("esios_public", None, "2032-10-26", 1, 200, ClientError),
    ),
)
@pytest.mark.asyncio
async def test_bad_downloads(
    data_source,
    api_token,
    day_str,
    num_log_msgs,
    status,
    exception,
    caplog,
):
    """Test data parsing of official API files."""
    day = datetime.fromisoformat(day_str).astimezone(REFERENCE_TZ)
    mock_session = MockAsyncSession(status=status, exc=exception)
    with caplog.at_level(logging.INFO):
        pvpc_data = PVPCData(
            session=mock_session,
            data_source=cast(DataSource, data_source),
            api_token=api_token,
        )
        if status in (401, 403):
            with pytest.raises(BadApiTokenAuthError):
                await pvpc_data.async_update_all(None, day)
            assert mock_session.call_count == 1
            return

        api_data = await pvpc_data.async_update_all(None, day)
        assert not api_data.sensors[KEY_PVPC]
        assert not pvpc_data.process_state_and_attributes(api_data, KEY_PVPC, day)
        assert len(caplog.messages) == num_log_msgs
    assert mock_session.call_count == 1
    check_num_datapoints(api_data, (KEY_PVPC,), 0)


# TODO review download schedule for Canary Islands TZ
@pytest.mark.parametrize(
    "local_tz, data_source, sensor_keys",
    (
        (TZ_TEST, "esios_public", (KEY_PVPC,)),
        (REFERENCE_TZ, "esios_public", (KEY_PVPC,)),
        (TZ_TEST, "esios", (KEY_PVPC, KEY_INJECTION, KEY_OMIE)),
        (REFERENCE_TZ, "esios", (KEY_PVPC, KEY_INJECTION, KEY_OMIE)),
    ),
)
@pytest.mark.asyncio
async def test_reduced_api_download_rate_dst_change(local_tz, data_source, sensor_keys):
    """Test time evolution and number of API calls."""
    start = datetime(2021, 10, 30, 15, tzinfo=UTC_TZ)
    mock_session = MockAsyncSession()
    pvpc_data = PVPCData(
        session=mock_session,
        tariff="2.0TD",
        local_timezone=local_tz,
        data_source=cast(DataSource, data_source),
        api_token="test-token" if data_source == "esios" else None,
        sensor_keys=sensor_keys,
    )
    assert pvpc_data.attribution == ATTRIBUTIONS[data_source]

    # avoid extra calls at day if already got all today prices
    api_data = None
    for _ in range(3):
        start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
        assert mock_session.call_count == len(sensor_keys)
        check_num_datapoints(api_data, sensor_keys, 24)
    assert all(api_data.availability.values())

    # first call for next-day prices
    assert start == datetime(2021, 10, 30, 18, tzinfo=UTC_TZ)
    start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
    assert mock_session.call_count == 2 * len(sensor_keys)
    check_num_datapoints(api_data, sensor_keys, 49)

    # avoid calls at evening if already got all today+tomorrow prices
    for _ in range(3):
        start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
        assert mock_session.call_count == 2 * len(sensor_keys)
        check_num_datapoints(api_data, sensor_keys, 49)

    # avoid calls at day if already got all today prices
    for _ in range(21):
        start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
        assert mock_session.call_count == 2 * len(sensor_keys)
        assert all(api_data.availability.values())
        # check_num_datapoints(api_data, sensor_keys, 25)

    # call for next-day prices (no more available)
    assert start == datetime(2021, 10, 31, 19, tzinfo=UTC_TZ)
    call_count = mock_session.call_count
    while start.astimezone(local_tz) <= datetime(2021, 10, 31, 23, tzinfo=local_tz):
        start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
        call_count += len(sensor_keys)
        assert mock_session.call_count == call_count
        # check_num_datapoints(api_data, sensor_keys, 25)

    # assert mock_session.call_count == 6
    assert pvpc_data.states.get(KEY_PVPC)
    assert all(api_data.availability.values())
    assert start.astimezone(local_tz) == datetime(2021, 11, 1, tzinfo=local_tz)
    assert not pvpc_data.process_state_and_attributes(api_data, KEY_PVPC, start)

    # After known prices are exausted, the state is flagged as unavailable
    with pytest.raises(AssertionError):
        start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
    assert not any(api_data.availability.values())
    start += timedelta(hours=1)
    assert not pvpc_data.process_state_and_attributes(api_data, KEY_PVPC, start)


@pytest.mark.parametrize(
    "local_tz, data_source, sensor_keys",
    (
        (TZ_TEST, "esios", ALL_SENSORS),
        (REFERENCE_TZ, "esios", ALL_SENSORS),
    ),
)
@pytest.mark.asyncio
async def test_reduced_api_download_rate(local_tz, data_source, sensor_keys):
    """Test time evolution and number of API calls."""
    start = datetime(2024, 3, 9, 2, tzinfo=UTC_TZ)
    mock_session = MockAsyncSession()
    pvpc_data = PVPCData(
        session=mock_session,
        tariff="2.0TD",
        local_timezone=local_tz,
        data_source=cast(DataSource, data_source),
        api_token="test-token" if data_source == "esios" else None,
        sensor_keys=sensor_keys,
    )
    assert pvpc_data.attribution == ATTRIBUTIONS[data_source]

    # avoid extra calls at day if already got all today prices
    api_data = None
    for _ in range(17):
        start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
        assert mock_session.call_count == len(sensor_keys)
        check_num_datapoints(api_data, sensor_keys, 24)

    # first call for next-day prices
    assert start == datetime(2024, 3, 9, 19, tzinfo=UTC_TZ)
    start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
    assert mock_session.call_count == 2 * len(sensor_keys)
    check_num_datapoints(api_data, sensor_keys, 24)

    call_count = mock_session.call_count
    while start.astimezone(local_tz) <= datetime(2024, 3, 9, 23, tzinfo=local_tz):
        start, api_data = await run_h_step(mock_session, pvpc_data, api_data, start)
        call_count += len(sensor_keys)
        assert mock_session.call_count == call_count
        check_num_datapoints(api_data, sensor_keys, 24)
